
/*****************************************************************
 * Dummyload projekt 2023 - 25 OZ7SKB Skanderborg EDR
 * 500 Watt udgave
 * 
 * Rev. 1.1 16052024  OZ4KAJ
 *      2.0 14122024  OZ9KHA Tilføjet auto skalering 0->100 eller 0->500 Watt, styre med FET og ekstra trimpot i OpAmp kredsløbet
 *      2.1 22012025  OZ4KAJ Refactoring
 *
 * Inspireret af W8TEE Jack Purdum og W6DQ Dennis Kidder
 * Arduino Projects for Amateur Radio 2015
 *
 * <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  5 linjer skal bruger tilrettes
 * 
 * !!!!!!!!!!!!!!!!! ADVARSEL !!!!!!!!!!!!!!!!!
 * Husk at koden skal kompileres og uploades til Arduino Uno, 
 * hvor kun Arduino Uno er tilkoblet. 
 * Alså uden display og board med elektronik tilkoblet.
 * RISIKO: Arduino Uno kan blive ødelagt med kortslutning af AREF-porten
 *****************************************************************/
 
// Inkluder hjælpelibrary
#include <LiquidCrystal.h>

// Konstanter og variabler defineres

#define STARTUPMSG "OZ7SKB Dummyload" //kan ændres til f.eks dit call - blot husk max 16 tegn <<<<<<<<<<<<<

#define DEBUG                   // Udmarkers med // til drift  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
#define DIODEVOLTAGEDROP  1.8   // Tre 1N4814...måske en anden værdi ved andre dioder <<<<<<<<<<<<<<<<<<<
#define UPDATEDELAY       100   // Pause til aflæsning af A1
#define ERRORDELAY        400   // Pause til visning af fejl
#define STARTUPDELAY     5000   // Pause ved opstart til visning af version mm

typedef enum {
  LowPower,
  HighPower
} RangeMode;

RangeMode range;

const int LCDNUMCOLS =     16;  // Antal kolonner på display
const int DLSENSORPIN =    A1;  // Definere analog port A1 som input
const int RANGEPIN =       8;   // On/off til FET der styrer high/low mode

const float MESASUREDDUMMYLOADRESISTANCE = 50.0;  // Modstanden i dummyload  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const int localVref = 3740;     // Vref målt til 3740 mV   <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
unsigned long holdTime;
int Wmax, cnt = 0;

// Definere LCD pins
const int rs = 12, en = 11, d4 = 5, d5 = 4, d6 = 3, d7 = 2;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);


/*********** Initialisering af Uno ***********/
void setup() {
  Serial.begin(9600);             // Initiere seriel kommunikation

  #ifdef DEBUGx
    delay(1000); // vent et øjeblik, så seriel kommunikationen er klar
    Serial.print("Compiler version: ");
    Serial.println(__VERSION__);
    Serial.print("Source fil: ");
    Serial.println(__FILE__);
    Serial.print("Compile date: ");
    Serial.println(__DATE__);
    Serial.print("Compile time: ");
    Serial.println(__TIME__);
  #endif
  pinMode(RANGEPIN, OUTPUT);      // set pin der styrer FET'en til output

  analogReference(EXTERNAL);      // Brug den eksterne spaendingsreference
  lcd.begin(16, 2);               // Initiere display til 16 x 2

  // Startup visning på display. Kan frit ændres. Max 16 x 2 tegn
  lcd.print(STARTUPMSG);    // Viser teksten
  lcd.setCursor(0, 1);              // Linje 2
  lcd.print("Version 2.1 500W");    // Viser teksten
  delay(STARTUPDELAY);              // Hold en pause
  lcd.clear();                      // Display renses  

  range = LowPower;                 // initieres til 100 Watt
  setRange(range);

} // setup() slut

/************ Arbejdsprogrammet gentages uendeligt *********/

void loop() {
  // Lokale variable
  float val, Vp, Vrms, Vin;
  int Watt;

  val = getAnalogValue();

  if (val > 3590 && range == HighPower) { // ved "meget høj" måling i high power mode har vi nået max effekt for dummy load'en - vis overload advarsel
      lcd.clear();
      lcd.print(" !! OVERLOAD !!");
      lcd.noDisplay();                   // flash the display on and off 
      delay(ERRORDELAY);
      lcd.display();
      delay(ERRORDELAY);   
      return; // restart loop
  }

  if (val > 3400 && range == LowPower) { // ved low power mode og vi får en "høj" måling, skiftes til high power mode
    // gå til HighPower mode
    range = HighPower;
    setRange(range);
    val = getAnalogValue(); // genindlæg analog værdi, nu hvor opamp forstærkning er justeret til HighPower
  }

  if (val < 1000 && range == HighPower) { // ved high power mode og vi får en "lav" målig kan vi gå tilbage til low power mode
    // gå til LowPower mode
    range = LowPower;
    setRange(range);
    val = getAnalogValue(); // genindlæg analog værdi, nu hvor opamp forstærkning er justeret til LowPower
  }

  if (range == LowPower) {
    Vin = (val) / (3600/51); // omregning så 0 -> Vref kommer til at svare til input 0 -> 51mV
  } else {
    Vin = (val) / (3600/114); // omregning så 0 -> Vref kommer til at svare til input 0 -> 114mV
  }

  Vp = Vin*100051/51000; // beregning af Vp på dummy load - spændingstab over 100Kohm i serie med 51ohm. Yderligere divideres med 1000 så mV omregnes til Volt
  Vp = Vp + 2*DIODEVOLTAGEDROP; // voltage drop over 2 x dioder
  Vrms = Vp/1.414; // beregn Vrms
  Watt = (Vrms * Vrms) / MESASUREDDUMMYLOADRESISTANCE; // så har vi Watt afsat i dummy load'en

  // måling og nulstilning af den max målte effekt
  if (Wmax < Watt) Wmax = Watt; // husk højest målte værdi
  if (Watt < 1 && Wmax > 0) {  // hvis vi ikke måler nogen værdi mere og vi har haft målt en max, vil vi nulstille max efter ca. 3 sekunder
    if (holdTime == 0) holdTime = millis();
    if (millis() - holdTime > 3000) {
      Wmax = 0;
      holdTime = 0;
    } 
  }

  // vis måle værdier i LCD display
  if (cnt-- == 0) { //opdater kun tallene i display f.eks hver sekund for at undgå "flicker"
    showWatt(Watt, Wmax);
    cnt = 1000/UPDATEDELAY; // giver opdatering af LCD hver sekund (så længe UPDATEDELAY er mindre en 1000)
  }
  showBar(range, Watt); // opdater bar hver gang vi har en aflæsning

  delay(UPDATEDELAY);
} //loop() slut

/***************** Funktioner og procedurer ***********************/
 

void setRange(RangeMode range) {
  // opsætning af low eller high power mode
  // i low mode måles fra 0 -> 100 Watt
  // i high mode måles fra 0 -> 500 Watt
  // FET styrer forstærkningen på OpAmp
  lcd.setCursor(12, 1);
  if (range == LowPower) {
    // FET on
                                                // opamp gain trimmes til 72, så 50 mV bliver til full scale ???
    digitalWrite(RANGEPIN, 1);
    lcd.print(" LOW");
  } else {
    // FET off
                                              // opamp gain trimmes til 32, så 111mV bliver full scale   ???
    digitalWrite(RANGEPIN, 0);
    lcd.print("HIGH");
  }
} // setRange() slut


void showWatt(int Watt, int Wmax) {
  // vis aktuel og max måling på linje 1 i LCD display
  char buffer[LCDNUMCOLS + 1];    // Buffer til output
  lcd.setCursor(0,0);             // Placerer cursor i 0,0 på display
  sprintf(buffer,"   %3dW   (%3dW)", Watt, Wmax);  // skriv output til buffer
  lcd.print(buffer);  
} //showWatt() slut


void showBar(RangeMode range, float power) 
{
  // vis bar på 10 tegn på linje 2 i LCD display
  // ved low power viser hver tegn 10 Watt ved high power 50 Watt
  int n;
  for (int i = 0; i < 10; i++) 
  {
      lcd.setCursor(i, 1);
      if ((range == LowPower && power > (10*i) + 1) || (range == HighPower && power > (50*i) + 1))
      {
        lcd.print((char) 255);
      } 
      else 
      {
        lcd.print(" ");
      }
      //Wmax = 480;         //************************** test  ***********************
      if (range == LowPower) n = Wmax / 10;
        else n = Wmax / 50;
        lcd.setCursor(n  , 1);
        lcd.print((char) 127);
        
  }
} // showBar() slut


int getAnalogValue() {
  // returnerer en værdi mellem 0 og localVref
  int val;
  val = map(analogRead(DLSENSORPIN), 0, 1024, 0, localVref); // input 0 -> 1024 mappes til spændingen 0 -> Vref (målt til 3600mV)
  // til test ******************************************************************************
  //val = map(990, 0, 1024, 0, localVref); // input 0 -> 1024 mappes til spændingen 0 -> Vref (målt til 3600mV)
  #ifdef DEBUG
    Serial.print("Analog value: ");
    Serial.println(val);
  #endif
  return val;
} 


